import { filter, map, Observable, Observer, Subject, Subscription } from "rxjs"
import { v4 as uuidv4 } from 'uuid'
import ConsoleLogger from "../utils/log.utils"
import { AdapterInterface, ClientObject, GeneralEvent, TransmissionRole, TransportType, TransportServiceInterface } from "../interface/interface"
import { TransmitterAdapter } from "./adapter.transmitter";
import { ReceiverAdapter } from "./adapter.receiver";
import { AdapterManagerBase } from "../base/adapter.manager.base";

export class AdapterManager extends AdapterManagerBase {
    private console: ConsoleLogger = new ConsoleLogger(`AdapterManager`, ['managers'])

    constructor(generalEvent: Subject<GeneralEvent<any>>, browserEnv?: boolean) {
        super()
        this.browserEnv = browserEnv ?? false
        this.console.log({ message: `Contructing self... ${this.browserEnv ? `is receiving end` : `is not browser env`}` })
        this.event = new Subject<GeneralEvent<any>>()

        this.connectToExistingTransport(generalEvent)
        this.initializeAdapterInstantiation(generalEvent, this.event)
    }

    public subscribeForAdapters(clientId: string, role: TransmissionRole): Observable<AdapterInterface> {
        return new Observable((adapters: Observer<AdapterInterface>) => {
            this.adapters.forEach((adapter: AdapterInterface) => {
                if (adapter.getAdapterProfile(`clientId`) == clientId && adapter.getAdapterProfile(`role`) == role) {
                    adapters.next(adapter)
                }
            })

            // According to chatgpt, any new adpaters that I insert into the array therafter will not be picked up, so need to manually listen to adapter evnet
            const subscription: Subscription = this.event.pipe(
                filter(event => event.type === `Adapter Event`),
                filter(event => event.event === `New Adapter`),
                filter(event => (event.data as AdapterInterface).getAdapterProfile(`role`) === role),
                filter(event => (event.data as AdapterInterface).getAdapterProfile(`clientId`) === clientId),
                map(event => { return (event.data as AdapterInterface) })
            ).subscribe(adapter => {
                adapters.next(adapter)
            })

            return () => {
                subscription.unsubscribe()
            }
        })
    }

    private initializeAdapterInstantiation(generalEvent: Subject<GeneralEvent<any>>, transportEvent: Subject<GeneralEvent<any>>): void {
        generalEvent.pipe(
            filter(event => event.type === `Transport Event`),
            map(event => { return event as GeneralEvent<any> })
        ).subscribe(transportEvent)

        transportEvent.pipe(
            filter(event => event.type === `Transport Event`),
        ).subscribe(event => {
            if (event.event == `New Client` || event.event == `New Server`) {
                let adapters: AdapterInterface[] = this.instantiateAdapterComponents(event.data as ClientObject, event.transport, event.transportServiceId as string)
                adapters.forEach(adapter => {
                    transportEvent.next({
                        id: uuidv4(),
                        type: `Adapter Event`,
                        event: `New Adapter`,
                        date: new Date(),
                        data: adapter,
                        transport: event.transport,
                        transportServiceId: event.transportServiceId
                    } as GeneralEvent<AdapterInterface>)
                })
            } else if (event.event == `Client Disconnected` || event.event == `Server Disconnected`) {
                // logic here ... pass...
            }
        })
    }

    private connectToExistingTransport(generalEvent: Subject<GeneralEvent<any>>): void {
        generalEvent.pipe(
            filter(event => event.type === `General Event`),
            filter(event => event.event === `Available Transport` || event.event === `New Transport`)
        ).subscribe((event: GeneralEvent<TransportServiceInterface[]>) => {
            let transportServices: TransportServiceInterface[] = event.data
            transportServices.forEach((transportService: TransportServiceInterface) => {
                this.console.log({ message: `updating transport in adapter ${transportService.getInfo().transportServiceId}` })
                this.updateTransportServicesRecord(transportService)
            })
        })
        // to automatically connect to existing started transport services
        generalEvent.next({
            id: uuidv4(),
            type: 'Adapter Event',
            event: `Adapter Manager Started`,
            date: new Date(),
            data: null
        } as GeneralEvent<any>)
    }

    private updateTransportServicesRecord(transportService: TransportServiceInterface): void {
        if (this.transportServiceArray.some(obj => obj.getInfo().transportServiceId === transportService.getInfo().transportServiceId)) {
            this.console.error({ message: `Transport Service Id: ${transportService.getInfo().transportServiceId} already existed.` })
        } else {
            this.transportServiceArray.push(transportService)
            this.console.log({ message: `Transport Service Id: ${transportService.getInfo().transportServiceId} has been added to adapter's record.` })
        }
    }
    /* At the moment, all adapters are instantiated based on which transport the client is connected. For multi type adapters instantiation logic for one client, this 
    will be for future enhancements. */
    private instantiateAdapterComponents(clientRef: ClientObject, transportType: TransportType, transportServiceId: string): AdapterInterface[] {
        let transportService = this.transportServiceArray.find(obj => obj.getInfo().transportServiceId === transportServiceId)
        if (transportService) {
            this.console.log({ message: `Instantiating ${transportType} adapters for client ${clientRef.clientId}` });
            const adapters: AdapterInterface[] = [
                new TransmitterAdapter(clientRef.clientId, transportService.getInfo().transport, transportService),
                new ReceiverAdapter(clientRef.clientId, transportService.getInfo().transport, transportService),
            ];
            this.adapters.concat(adapters)
            return adapters;
        } else {
            this.console.error({ message: `Transport Service id ${transportServiceId} not found. Unable to instantiate adapters.` })
            return []
        }
    }
}


