import dotenv from 'dotenv';
import { BehaviorSubject, distinctUntilChanged, filter, map, Observable, Observer, Subscription, takeWhile } from 'rxjs';
import { v4 as uuidv4 } from 'uuid'
import ConsoleLogger from '../utils/log.utils';
import { WrappedMessage } from '../utils/message.ordering';
import { AdapterBase } from '../base/adapter.base';
import { ClientObject, ConnectionState, FisMessage, GeneralEvent, TransportType, TransportMessage, TransportServiceInterface } from '../interface/interface';

dotenv.config();
/* This transport manager will be instantiating the necessary transport to deal with tranmission and receiving from different receivers
So how?: */
export class ReceiverAdapter extends AdapterBase {
    private console!: ConsoleLogger

    constructor(clientId: string, adapterType: TransportType, transportService: TransportServiceInterface) {
        super()
        this.console = new ConsoleLogger(`${adapterType}ReceiverAdapter`, ['adapter'])
        this.setAdapterProfile(clientId, adapterType, transportService, 'Receiver')
        this.setupConnectionState(transportService)

        this.console.log({ message: `Contructing ReceiverAdapter for clientId: ${clientId}` })
    }

    subscribeForIncoming(): Observable<GeneralEvent<any>> {
        this.console.log({ message: `Connector getting message bus for this connector: ${this.AdapterProfile.clientId}` })
        return new Observable((observable: Observer<GeneralEvent<any>>) => {
            const subscription: Subscription = this.AdapterProfile.transportService.subscribeForEvent().pipe(
                filter(event => event.type === `Transport Event`),
                filter((message: GeneralEvent<TransportMessage>) => (message.data as TransportMessage).self === this.AdapterProfile.clientId ),
                filter((message: GeneralEvent<any>) => message.event === 'New Message'),
            ).subscribe((message: GeneralEvent<TransportMessage>) => {
                this.console.log({ message: `Received ${(((message.data as TransportMessage).payload as WrappedMessage).payload as FisMessage).header.messageID} from ${((message.data as TransportMessage).target)}`, details: message })
                observable.next({
                    id: uuidv4(),
                    type: `Adapter Event`,
                    event: `New Message`,
                    date: new Date(),
                    data: message.data,
                    transport: undefined,
                    transportServiceId: null
                } as GeneralEvent<TransportMessage>);
            });

            // Clean up on unsubscription
            return () => {
                subscription.unsubscribe();
            };

        });
    }

    // generally not required but can be used for other indicative purposes
    getConnectionState(): Observable<ConnectionState> {
        return this.AdapterProfile.connectionState.asObservable()
    }

    // this is irrelevant at this point in time. Adapter will just listen regardless of whether there's connection or not
    private setupConnectionState(transportService: TransportServiceInterface): void {
        transportService.subscribeForEvent().pipe(
            filter(event => event.type === `Transport Event`),
            filter(event => (event.data as ClientObject).clientId === this.AdapterProfile.clientId),
            map(event => {
                if (event.event == 'Client Disconnected' || event.event == 'Server Disconnected') {
                    return 'OFFLINE'
                } else {
                    return `ONLINE`
                }
            }),
            distinctUntilChanged()
        ).subscribe((signal: ConnectionState) => {
            this.AdapterProfile.connectionState.next(signal)
            if (signal == 'OFFLINE') this.console.error({ message: `${this.AdapterProfile.clientId} disconnected` })
            if (signal == 'ONLINE') this.console.log({ message: `${this.AdapterProfile.clientId} connected and ready to go` })
        })
    }


}


