import { filter, interval, map, Observable, Observer, Subject, Subscription } from "rxjs";
import { v4 as uuidv4 } from 'uuid'
import config from '../config/config.json';
import { MessageTransmissionManager } from "../transmission/msg.transmission.manager";
import { WrappedMessage } from "../utils/message.ordering";
import ConsoleLogger from "../utils/log.utils";
import { FisMessage, GeneralEvent, Transport, TransportMessage, TransportServiceInterface, TransportSet } from "../interface/interface";
import { WebsocketTransportService } from "../transport/websocket";
import { HttpTransportService } from "../transport/http";
import clientProfile from '../../clients/serverprofile.json'
import serverProfile from '../../clients/serverprofile.json'
import { MessageTransmissionTransmitter } from "../transmission/msg.transmission.transmitter";
import { MessageTransmissionReceiver } from "../transmission/msg.transmission.receiver";
import { MessageTransmissionRequestResponse } from "../transmission/msg.transmission.request-response";

/* This is Emulating UI perspective. Using just browser environment, and not using any dotenv. */
class Supervisor {
    private generalBus: Subject<GeneralEvent<any>> = new Subject()
    private console = new ConsoleLogger('Supervisor', ['base'])
    private isClient: boolean = true
    private transmissionManager!: MessageTransmissionManager
    private event: Subject<GeneralEvent<any>>
    private outgoingPipe: Subject<any> = new Subject()
    private transportServiceArray: TransportServiceInterface[] = []
    private transportSet: TransportSet[] = []

    constructor() {
        this.event = new Subject()
        // Start setting up existing transport based on .env file
        this.sortTransportFromEnv(this.transportSet)
        this.transportSet.forEach(transport => {
            this.setUpTransportService(transport, this.event, this.isClient)
        })
        // once adapter manager is instantiated, it will attempt to connect to existing started transport
        this.tieInAdapterWithExistingTransportServices(this.event)

        this.transmissionManager = new MessageTransmissionManager(this.event, this.isClient)
        this.startMessageTransmission()
    }

    private startMessageTransmission(): void {
        let transmitter: MessageTransmissionTransmitter = this.transmissionManager.getTransmitter(clientProfile.clientId) as MessageTransmissionTransmitter
        let receiver: MessageTransmissionReceiver = this.transmissionManager.getReceiver(clientProfile.clientId) as MessageTransmissionReceiver
        let requestResponse: MessageTransmissionRequestResponse = this.transmissionManager.getRequestResponse(clientProfile.id, transmitter, receiver)

        // emit Message only
        // this.emitMessage(transmitter, this.generateNotifcation())

        // receive Message only
        this.streamMessage(receiver)

        // request-response emulation
        // this.sendMessage(requestResponse, { header: { messageID: `123`, messageName: `RequestMessage` }, data: `Test Data` } as FisMessage).subscribe({
        //     next: message => this.console.log({ message: `Received response message for request ${(message as FisMessage).header.messageID}` }),
        //     error: error => this.console.error({ message: `Something went wrong`, details: error }),
        //     complete: () => this.console.log({ message: `Request completed` })
        // })
    }


    private emitMessage(transmitter: MessageTransmissionTransmitter, source: Subject<any> | Observable<any>): void {
        source.subscribe(message => transmitter.emit(message))
    }

    private sendMessage(requestResponse: MessageTransmissionRequestResponse, message: any): Observable<any> {
        return new Observable((response: Observer<any>) => {
            requestResponse.send(message).subscribe({
                next: respMessage => {
                    response.next(respMessage)
                },
                error: error => {
                    this.console.error({ message: error })
                    response.error(error)
                },
                complete: () => {
                    this.console.log({ message: `Response Completed for message: ${message.id ?? 'undefined message'}` })
                    response.complete()
                }
            })
        })
    }

    private streamMessage(receiverInstance: MessageTransmissionReceiver): Observable<any> {
        return new Observable((response: Observer<any>) => {
            const subscription: Subscription = receiverInstance.getReceivables().subscribe({
                next: message => {
                    response.next(message)
                },
                error: error => response.error(error),
                complete: () => response.complete()
            })

            // Clean up on unsubscription
            return () => {
                subscription.unsubscribe();
            };
        })
    }

    private setUpTransportService(transportSet: TransportSet, event: Subject<GeneralEvent<any>>, isClient?: boolean): void {
        try {
            let transportService: TransportServiceInterface = this.instantiateTransportService(transportSet.transport, event)
            this.transportServiceArray.push(transportService)
            if (transportService instanceof WebsocketTransportService) {
                this.console.log({ message: `Just Double Checking... this is websocket` })
                if (isClient) {
                    // please note this is subject to change depending on the UI environemnt. Angular has their own built in function to read json file based on Swopt-UI
                    transportService.startClient(config.connection.transmitter)
                } else {
                    transportService.startServer(transportSet.port);
                }
            } else if (transportService instanceof HttpTransportService) {
                this.console.log({ message: `Just Double Checking... this is http` })
                // Additional Http-specific setup if needed.
                if (isClient) {
                    transportService.startClient(config.connection.transmitter)
                } else {
                    transportService.startServer(transportSet.port)
                }
            }
        } catch (error) {
            this.console.error({ message: 'Fail to set transport. Error in setting up transport', details: error })
        }
    }

    private instantiateTransportService(transportType: Transport, event: Subject<GeneralEvent<any>>): TransportServiceInterface {
        if (transportType === 'Websocket') {
            return new WebsocketTransportService(event)
        }
        else if (transportType === 'Http') {
            return new HttpTransportService(event)
        } else {
            throw new Error(`No Transport Service Instantiated`)
        }
    }

    // just to re-arrange the list of transport servicce from env. Of course, varying modules will have varying ways to start up their transport respectively. TO adapter later
    private sortTransportFromEnv(transportSet: TransportSet[]): void {
        let transportList: string[] = process.env.Transport?.split(',') || []
        let portList: number[] = (process.env.PORT?.split(',') || []).map(port => Number(port));
        transportList.forEach((transport, index) => {
            transportSet.push({ transport: transport, port: portList[index] } as unknown as TransportSet)
        })
        this.console.log({ message: 'TransportSetList', details: this.transportSet })
    }

    // A method to broadcast avaible tranpsort to adapterManager the available transport to be connected after the adapter manager has been instantiated
    private tieInAdapterWithExistingTransportServices(eventBus: Subject<GeneralEvent<any>>): void {
        const subscription: Subscription = eventBus.pipe(
            filter(event => event.type === `Adapter Event`),
            filter(event => event.event === `Adapter Manager Started`)
        ).subscribe(event => {
            eventBus.next({
                id: uuidv4(),
                type: `General Event`,
                event: `Available Transport`,
                date: new Date(),
                data: this.transportServiceArray,
                transport: undefined,
                transportServiceId: null
            })
        })
    }

    private generateNotifcation(): Observable<FisMessage> {
        return new Observable((response: Observer<FisMessage>) => {
            const intervalMessageGeneration = interval(1000).pipe(
                map(() => {
                    const message: FisMessage = {
                        header: {
                            messageID: uuidv4(),
                            messageName: 'NotificationMessage'
                        },
                        data: `Data`
                    };
                    return message;
                })
            );

            const subscription = intervalMessageGeneration.subscribe({
                next: message => response.next(message),
                error: error => response.error(error),
                complete: () => {
                    response.next({
                        header: {
                            messageID: uuidv4(),
                            messageName: 'NotificationMessage'
                        },
                        data: `Complete`
                    });
                    response.complete();
                }
            });

            // Ensure cleanup on unsubscribe
            return () => subscription.unsubscribe();
        });
    }

}

let supervisor = new Supervisor()
