import { filter, interval, map, Observable, Observer, Subject, Subscription, take } from "rxjs";
import { v4 as uuidv4 } from 'uuid'
import { MessageTransmissionManager } from "../transmission/msg.transmission.manager";
import { WrappedMessage } from "../utils/message.ordering";
import ConsoleLogger from "../utils/log.utils";
import { FisMessage, GeneralEvent, MessageTransmissionInterface, Transport, TransportMessage, TransportServiceInterface, TransportSet } from "../interface/interface";
import config from '../config/config.json';
import { startSocketServer } from "../utils/socket.utils";
import { WebsocketTransportService } from "../transport/websocket";
import { HttpTransportService } from "../transport/http";
import clientProfile from '../../clients/serverprofile.json'
import serverProfile from '../../clients/serverprofile.json'
import { MessageTransmissionReceiver } from "../transmission/msg.transmission.receiver";
import { MessageTransmissionTransmitter } from "../transmission/msg.transmission.transmitter";
import { MessageTransmissionRequestResponse } from "../transmission/msg.transmission.request-response";
import { error } from "console";
class Supervisor {
    private console = new ConsoleLogger('Supervisor', ['base'])
    private clientIncomingMessage: Subject<FisMessage> = new Subject()
    private messageProducer!: MessageProducer
    private transmissionManager!: MessageTransmissionManager
    private event!: Subject<GeneralEvent<any>>
    private transportSet: TransportSet[] = []
    private transportServiceArray: TransportServiceInterface[] = []

    constructor() {
        this.event = new Subject<GeneralEvent<any>>()
        // set up all the transportServices
        this.sortTransportFromEnv(this.transportSet)
        this.transportSet.forEach(set => {
            this.setUpTransportService(set, this.event)
        })
        this.tieInAdapterWithExistingTransportServices(this.event)
        // so need them adapters now. But supervisor shouldn't be concerned, only messageTransmissionManager and ConnectionManager
        this.messageProducer = new MessageProducer(this.clientIncomingMessage)
        this.transmissionManager = new MessageTransmissionManager(this.event)
        this.startMessageTransmission()

    }

    private startMessageTransmission(): void {
        let transmitter: MessageTransmissionTransmitter = this.transmissionManager.getTransmitter(serverProfile.clientId) as MessageTransmissionTransmitter
        let receiver: MessageTransmissionReceiver = this.transmissionManager.getReceiver(serverProfile.clientId) as MessageTransmissionReceiver
        let requestResponse: MessageTransmissionRequestResponse = this.transmissionManager.getRequestResponse(serverProfile.id, transmitter, receiver)

        // emit Message only
        this.emitMessage(transmitter, this.messageProducer.getNotificationMessage())
        // this.emitMessage(transmitter, this.messageProducer.getOutgoingMessages())

        // receive Message only
        this.streamMessage(receiver).subscribe(message => {
            this.console.log({ message: message.header.messageID })
            // this.clientIncomingMessage.next(message)
        })

        // request-response emulation
        this.sendMessage(requestResponse, { header: { messageID: `123`, messageName: `RequestMessage` }, data: `Test Data` } as FisMessage).subscribe({
            next: message => this.console.log({ message: `Received response message for request ${(message as FisMessage).header.messageID}` }),
            error: error => this.console.error({ message: `Something went wrong`, details: error }),
            complete: () => this.console.log({ message: `Request completed` })
        })
    }

    private emitMessage(transmitter: MessageTransmissionTransmitter, source: Subject<any> | Observable<any>): void {
        source.subscribe(message => transmitter.emit(message))
    }

    private sendMessage(requestResponse: MessageTransmissionRequestResponse, message: any): Observable<any> {
        return new Observable((response: Observer<any>) => {
            requestResponse.send(message).subscribe({
                next: respMessage => {
                    response.next(respMessage)
                },
                error: error => {
                    this.console.error({ message: error })
                    response.error(error)
                },
                complete: () => {
                    this.console.log({ message: `Response Completed for message: ${message.id ?? 'undefined message'}` })
                    response.complete()
                }
            })
        })
    }

    private streamMessage(receiverInstance: MessageTransmissionReceiver): Observable<any> {
        return new Observable((response: Observer<any>) => {
            const subscription: Subscription = receiverInstance.getReceivables().subscribe({
                next: message => {
                    response.next(message)
                },
                error: error => response.error(error),
                complete: () => response.complete()
            })

            // Clean up on unsubscription
            return () => {
                subscription.unsubscribe();
            };
        })
    }

    // Server to be set up as well as acquiring client information if needed. Like in the case for grpc and socket. Http not requ`ired.
    private setUpTransportService(transportSet: TransportSet, event: Subject<GeneralEvent<any>>, isClient?: boolean): void {
        try {
            let transportService: TransportServiceInterface = this.instantiateTransportService(transportSet.transport, event)
            this.transportServiceArray.push(transportService)
            if (transportService instanceof WebsocketTransportService) {
                this.console.log({ message: `Just Double Checking... this is websocket` })
                if (isClient) {
                    // please note this is subject to change depending on the UI environemnt. Angular has their own built in function to read json file based on Swopt-UI
                    transportService.startClient(config.connection.transmitter)
                } else {
                    transportService.startServer(transportSet.port);
                }
            } else if (transportService instanceof HttpTransportService) {
                this.console.log({ message: `Just Double Checking... this is http` })
                // Additional Http-specific setup if needed.
                if (isClient) {
                    transportService.startClient(config.connection.transmitter)
                } else {
                    transportService.startServer(transportSet.port)
                }
            }
        } catch (error) {
            this.console.error({ message: 'Fail to set transport. Error in setting up transport', details: error })
        }
    }

    private instantiateTransportService(transportType: Transport, event: Subject<GeneralEvent<any>>): TransportServiceInterface {
        if (transportType === 'Websocket') {
            return new WebsocketTransportService(event)
        }
        else if (transportType === 'Http') {
            return new HttpTransportService(event)
        } else {
            throw new Error(`No Transport Service Instantiated`)
        }
    }

    private sortTransportFromEnv(transportSet: TransportSet[]): void {
        let transportList: string[] = process.env.Transport?.split(',') || []
        let portList: number[] = (process.env.PORT?.split(',') || []).map(port => Number(port));
        transportList.forEach((transport, index) => {
            transportSet.push({ transport: transport, port: portList[index] } as unknown as TransportSet)
        })
        this.console.log({ message: 'TransportSetList', details: this.transportSet })
    }

    private tieInAdapterWithExistingTransportServices(eventBus: Subject<GeneralEvent<any>>): void {
        const subscription: Subscription = eventBus.pipe(
            filter(event => event.type === `Adapter Event`),
            filter(event => event.event === `Adapter Manager Started`)
        ).subscribe(event => {
            eventBus.next({
                id: uuidv4(),
                type: `General Event`,
                event: `Available Transport`,
                date: new Date(),
                data: this.transportServiceArray,
                transport: undefined,
                transportServiceId: null
            })
        })
    }
}


class MessageProducer {
    private console = new ConsoleLogger('Message Producer', ['base'])
    private generalNotification: Subject<FisMessage> = new Subject()
    private incomingMessageBus!: Subject<FisMessage>
    private outgoingMessageBus: Subject<FisMessage> = new Subject()

    constructor(incomingMessageBus: Subject<FisMessage>) {
        this.console.log({ message: `Constructing Message Producer` })
        this.incomingMessageBus = incomingMessageBus

        this.generateNotifcation().subscribe(this.generalNotification)
        this.handleIncomingRequests(this.incomingMessageBus.asObservable(), this.outgoingMessageBus)
    }

    public getNotificationMessage(): Observable<FisMessage> {
        return this.generalNotification.asObservable()
    }

    public getOutgoingMessages(): Observable<FisMessage> {
        return this.outgoingMessageBus.asObservable()
    }

    // this is called no problem
    private handleIncomingRequests(requests: Observable<FisMessage>, outgoingMessageBus: Subject<FisMessage>): void {
        requests.subscribe((request: FisMessage) => {
            this.console.log({ message: `Generating response for new request ${request.header.messageID}` })
            this.generateMessage(request.header.messageID, 10).subscribe({
                next: message => outgoingMessageBus.next(message),
                error: error => this.console.log({ message: 'observer Error', details: error }),
                complete: () => {
                    outgoingMessageBus.next({
                        header: {
                            messageID: request.header.messageID,
                            messageName: `ResponseMessage`
                        },
                        data: `Complete`
                    } as FisMessage)
                }
            })
        })
    }

    private generateMessage(requestID: string, amount: number): Observable<FisMessage> {
        return new Observable((response: Observer<FisMessage>) => {
            const intervalMessageGeneration = interval(1000).pipe(
                take(amount), // Ensures only 'amount' messages are generated
                map(() => {
                    const message: FisMessage = {
                        header: {
                            messageID: requestID,
                            messageName: 'ResponseMessage'
                        },
                        data: `Data`
                    };
                    return message;
                })
            );

            const subscription = intervalMessageGeneration.subscribe({
                next: message => response.next(message),
                error: error => response.error(error),
                complete: () => {
                    response.next({
                        header: {
                            messageID: requestID,
                            messageName: 'ResponseMessage'
                        },
                        data: `Complete`
                    });
                    response.complete();
                }
            });

            // Ensure cleanup on unsubscribe
            return () => subscription.unsubscribe();
        });
    }

    private generateNotifcation(): Observable<FisMessage> {
        return new Observable((response: Observer<FisMessage>) => {
            const intervalMessageGeneration = interval(1000).pipe(
                map(() => {
                    const message: FisMessage = {
                        header: {
                            messageID: uuidv4(),
                            messageName: 'NotificationMessage'
                        },
                        data: `Data`
                    };
                    return message;
                })
            );

            const subscription = intervalMessageGeneration.subscribe({
                next: message => response.next(message),
                error: error => response.error(error),
                complete: () => {
                    response.next({
                        header: {
                            messageID: uuidv4(),
                            messageName: 'NotificationMessage'
                        },
                        data: `Complete`
                    });
                    response.complete();
                }
            });

            // Ensure cleanup on unsubscribe
            return () => subscription.unsubscribe();
        });
    }
}

let supervisor = new Supervisor()
