import { MessageTransmissionBase } from "../base/msg.transmission.base";
import { v4 as uuidv4 } from 'uuid'
import { BehaviorSubject, distinctUntilChanged, filter, map, Observable, Subject, Subscription } from "rxjs";
import { RetransmissionService } from "../utils/retransmission.service";
import { WrappedMessage } from "../utils/message.ordering";
import ConsoleLogger from "../utils/log.utils";
import { AdapterInterface, AdapterManagerInterface, ConnectionState, FisMessage, GeneralEvent, MessageTransmitterInterface, TransmitterAdapterInterface, TransportMessage } from "../interface/interface";
import { error } from "console";
import { TransmitterAdapter } from "../adapters/adapter.transmitter";

/* Take in all the messages that needs to be transported, and divide them accordingly. So the connector instances will do just that
connectors or adapters will have their own identifier*/
export class MessageTransmissionTransmitter extends MessageTransmissionBase implements MessageTransmitterInterface {
    private connectionStateEvent: BehaviorSubject<ConnectionState> = new BehaviorSubject<ConnectionState>('OFFLINE')
    private console: ConsoleLogger = new ConsoleLogger(`MessageTransmissionTransmitter`, ['transmission'])
    private messageToBeBuffered!: Subject<FisMessage | WrappedMessage>
    private buffer!: RetransmissionService;
    private currentAdapter!: TransmitterAdapterInterface

    constructor(clientId: string, adapterManager: AdapterManagerInterface) {
        super()
        this.console.log({ message: `Constructing Transmitter Transmission with ${clientId}` })
        this.clientId = clientId
        this.messageToBeBuffered = new Subject()
        this.buffer = new RetransmissionService()
        this.initializeTransmitterComponents(adapterManager)
    }

    public emit(message: FisMessage): void {
        // this.console.log({ message: `${this.connectionStateEvent.getValue() == 'ONLINE' ? `Transmitting...` : `Buffering...`}` })
        this.messageToBeBuffered.next(message)
    }

    /* After setting up, will listen specifically to the connection state of this particular remote client. So that, the buffer signal can be
    established to allow the buffer to do their thing. */
    private initializeTransmitterComponents(adapterManager: AdapterManagerInterface): void {
        this.console.log({ message: `Setting up Retransmission Service...` })
        // Listen and update adapters
        adapterManager.subscribeForAdapters(this.clientId, `Transmitter`).subscribe((adapter: AdapterInterface) => {
            this.console.log({ message: `Adding new ${adapter.getAdapterProfile(`transportType`)} transmitting adapter. Current adapter length: ${this.adapters.length}` })
            this.adapters.push(adapter)
            if (!this.currentAdapter) {
                this.console.log({ message: `Setting this ${adapter.getAdapterProfile(`transportType`)} as current adapter.` })
                this.currentAdapter = adapter as TransmitterAdapterInterface
                this.currentAdapter.connectionState.subscribe(this.connectionStateEvent)
            }
        })

        this.buffer.implementRetransmission(this.messageToBeBuffered, this.connectionStateEvent.asObservable(), true)
        // automatically subscribe to allow released bffered messages to be released
        this.buffer.returnSubjectForBufferedItems().subscribe((bufferedMessage: WrappedMessage) => {
            // need to work with wrapped messages
            this.console.log({ message: `Transmitting ${bufferedMessage.thisMessageID}` });
            if (this.currentAdapter) {
                this.currentAdapter.emit(bufferedMessage)
            } else {
                // just flush back the message inside the buffer, if the adapter is not ready or assigned.
                this.messageToBeBuffered.next(bufferedMessage)
                this.console.error({ message: `Adapter is not set. Please ensure adapters are ready. Message ${(bufferedMessage.payload as FisMessage).header.messageID} is flushed back into buffer.` })
            }
        })
    }

    // this is for http only. Please ignore for now.
    private uniqueHandlerToFlushUnsentMessages(event: Observable<GeneralEvent<any>>): void {
        event.pipe(
            filter(event => event.event == 'Re-Flush'),
            filter(event => event.data.clientId == this.clientId),
        ).subscribe((event: GeneralEvent<any>) => {
            this.console.log({ message: `${this.connectionStateEvent.getValue() == 'ONLINE' ? `Transmitting ${((event.data.payload as TransportMessage).payload as WrappedMessage).thisMessageID}` : `Buffering ${((event.data.payload as TransportMessage).payload as WrappedMessage).thisMessageID}`}` })
            this.messageToBeBuffered.next(((event.data.payload as TransportMessage).payload as WrappedMessage))
        })
    }

}