import { Observable, Observer } from 'rxjs';
import { createServer } from 'http';
import { Server, Socket as SocketForConnectedClient } from 'socket.io';
import { io, Socket as SocketForConnectedServer } from 'socket.io-client';
import * as fs from 'fs'
import { v4 as uuidv4 } from 'uuid'
import { ConnectedSocketClient, ConnectedSocketServer } from '../transport/websocket';
import ConsoleLogger from './log.utils';
import path from 'path';
import { ClientObject, GeneralEvent, TransportMessage } from '../interface/interface';
import { addClientToDB, checkIfClientExists, checkOwnClientInfo } from './general.utils';
const console: ConsoleLogger = new ConsoleLogger(`SocketUtils`, ['transport'])

export function startSocketServer(port: number): Observable<SocketForConnectedClient> {
    return new Observable((observer) => {
        try {
            console.log({ message: `Socket Server ${port} Started....` })
            let httpServer = createServer();
            let socketServer = new Server(httpServer)

            // something wrong here
            socketServer.on('connection', (socket) => {
                observer.next(socket)
            })

            socketServer.engine.on("connection_error", (err) => {
                console.log({ message: `Socket Server ${port} Connection Error`, details: err.req })
                console.log({ message: `Socket Server ${port} Connection Error`, details: err.code })
                console.log({ message: `Socket Server ${port} Connection Error`, details: err.message })
                console.log({ message: `Socket Server ${port} Connection Error`, details: err.context })
            });

            // Start the HTTP server on 127.0.0.1 with the given port
            httpServer.listen(port, '0.0.0.0', () => {
                console.log({ message: `Socket server listening on ${port}` });
            });
        } catch (error) {
            observer.error(error);
        }
    })
}

export async function startClientSocketConnection(serverUrl: string): Promise<SocketForConnectedServer> {
    return new Promise((resolve, reject) => {
        try {
            // let clientSocket = io(serverUrl)
            let clientSocket: SocketForConnectedServer = io(serverUrl, {
                reconnection: true,              // Enable automatic reconnections
                reconnectionAttempts: 1000,       // Retry up to 10 times
                reconnectionDelay: 500,          // Start with a 500ms delay
                reconnectionDelayMax: 10000,     // Delay can grow to a max of 10 seconds
                randomizationFactor: 0.3,
            })
            resolve(clientSocket)
        }
        catch (error) {
            reject(error)
        }
    })
}

// After establishing connection to the server, set up the credentials, confirm whether or not if there's any credentials, if not ask for one from the server
export function handleClientSocketConnection(transportServiceId: string, socket: SocketForConnectedServer, serversConnected: ConnectedSocketServer[]): Observable<GeneralEvent<any>> {
    return new Observable((eventNotification: Observer<GeneralEvent<any>>) => {
        let buffer: any[] = []
        let selfId!: string
        let clientProfileInfo!: ConnectedSocketServer // this information refers to the server itself

        // Listen for a connection event
        socket.on('connect', () => {
            console.log({ message: `Connected to the server ${socket.id} ` })
            checkOwnClientInfo(`clientprofile`).then((profile: { id: string, clientId: string }) => {
                socket.emit('handshaking', {
                    id: profile.id,
                })
            }).catch((error) => {
                console.error({ message: `Client profile is not defined!`, details: error })
            })


            // Listen for messages from the server. Generally here's the responses
            socket.on('message', (msg: TransportMessage) => {
                if (clientProfileInfo) {
                    console.log({ message: `Websocket Client Transport Receive Msg` })
                    // publish to event
                    eventNotification.next({
                        id: uuidv4(),
                        type: `Transport Event`,
                        event: 'New Message',
                        date: new Date(),
                        data: msg,
                        transport: `Websocket`,
                        transportServiceId: transportServiceId
                    } as GeneralEvent<TransportMessage>)
                } else {
                    // Do nothing. just store in local array first. Cannot process without information. but then again, don['t need information if acting as client
                    // but for consistency sake, will impose the standard 
                    buffer.push(msg) // store locally for now
                }
            })

            socket.on('handshaking', (data: { id: string, message?: any }) => {
                // check if this is previously connected or not by checking if the clientId has been previously assigned
                if (serversConnected.some(obj => obj.clientId === data.id)) {
                    console.log({ message: data.message ?? `No message from server...` })
                    // Update websocket instance record
                    let clientObj: ConnectedSocketServer | undefined = serversConnected.find(obj => obj.clientId === data.id)
                    if (clientObj) {
                        console.log({
                            message: `Just to make sure they are pointed accurately !! Id match? ${clientProfileInfo.clientId == clientObj.clientId ? true : false} && compare ${clientObj.clientId}`,
                        })
                        // broadcast event to allow retransmission to release buffer
                        eventNotification.next({
                            id: uuidv4(),
                            type: `Transport Event`,
                            event: 'Server Connected',
                            date: new Date(),
                            data: clientObj,
                            transport: 'Websocket',
                            transportServiceId: transportServiceId
                        } as GeneralEvent<ClientObject>)
                    }
                } else {
                    console.log({ message: `Server acknowledged for new client connection for ${data.id}` })
                    clientProfileInfo = {
                        clientId: data.id,
                        dateCreated: new Date(),
                        socketInstance: socket
                    }
                    addClientToDB(clientProfileInfo as ConnectedSocketServer, 'servers')
                    /* Note that there are two separate events, because transmission must first be set up before releasing buffer. */
                    // broadcast event to allow transmission manager to instantiate transmission components
                    eventNotification.next({
                        id: uuidv4(),
                        type: 'Transport Event',
                        event: `New Server`,
                        date: new Date(),
                        data: clientProfileInfo,
                        transport: 'Websocket',
                        transportServiceId: transportServiceId
                    } as GeneralEvent<ClientObject>)
                    // broadcast event to allow retransmission to relase buffered messages
                    eventNotification.next({
                        id: uuidv4(),
                        type: 'Transport Event',
                        event: `Server Connected`,
                        date: new Date(),
                        data: clientProfileInfo,
                        transport: 'Websocket',
                        transportServiceId: transportServiceId
                    } as GeneralEvent<ClientObject>)
                    serversConnected.push(clientProfileInfo)
                }
            })

            // Handle disconnection
            socket.on('disconnect', () => {
                console.log({ message: `Socket Server ${clientProfileInfo.clientId} Disconnected` })
                if (clientProfileInfo) {
                    eventNotification.next({
                        id: uuidv4(),
                        type: `Transport Event`,
                        event: `Server Disconnected`,
                        date: new Date(),
                        data: clientProfileInfo,
                        transport: `Websocket`,
                        transportServiceId: transportServiceId
                    } as GeneralEvent<ClientObject>)
                }
            });
        })
    })
}

// For SERVER Usage: set up socket listeners to start listening for different events
export function handleNewSocketClient(transportServiceId: string, socket: SocketForConnectedClient, connectedClientSocket: ConnectedSocketClient[]): Observable<GeneralEvent<any>> {
    return new Observable((event: Observer<GeneralEvent<any>>) => {
        let selfId!: string
        checkOwnClientInfo(`serverprofile`).then((profile: { id: string }) => {
            selfId = profile.id
        })
        console.log({ message: `Socket client connected. Setting up listeners for socket:${socket.id}` })
        // returns the socket client instance 
        socket.on(`handshaking`, (data: { id: string, message?: any }) => {
            // if it's a previously connected client
            if (connectedClientSocket.some(obj => obj.clientId == data.id)) {
                // update first
                let clientInstance: ConnectedSocketClient | undefined
                if (connectedClientSocket.length > 0) {
                    clientInstance = connectedClientSocket.find(obj => obj.clientId === data.id)
                    handleFoundClient(clientInstance)
                } else {
                    // for the case server itself got shut down or something. This one will check at persisted local storage
                    checkIfClientExists(data.id).then((client: ConnectedSocketClient) => {
                        clientInstance = client
                        handleFoundClient(clientInstance)
                    }).catch(error => {
                        console.log({ message: `Promise Error`, details: error })
                    })
                }
                function handleFoundClient(clientInstance: ConnectedSocketClient | undefined) {
                    if (clientInstance) {
                        console.log({ message: `Socket Client ${clientInstance.clientId} Found. This is a previously connected client.` })
                        // publish own id to client
                        if (selfId) {
                            socket.emit('handshaking', { id: selfId })
                        } else {
                            console.error({ message: 'Self Id not established' })
                        }
                        // replace socket instance since the previous has been terminated
                        clientInstance.socketInstance = socket
                        // need to start listening again, because it's assigned a different socket instance this time round
                        startListening(socket, clientInstance, event, transportServiceId, true)
                    } else {
                        console.log({ message: `Profile Not Found` })
                        socket.emit('handshaking', { id: selfId, message: 'Receiver Profile Not found' })
                    }
                }
            } else {
                console.log({ message: `New Client Connected. ClientID: ${data.id}` })
                let clientInstance: ConnectedSocketClient = {
                    clientId: data.id,
                    dateCreated: new Date(),
                    socketInstance: socket,
                }
                // send to receiver for reference
                if (selfId) {
                    socket.emit('handshaking', { id: selfId })
                } else {
                    console.error({ message: 'Self Id not established' })
                }
                // publish first event notification
                event.next({
                    id: uuidv4(),
                    type: 'Transport Event',
                    event: `New Client`,
                    date: new Date(),
                    data: clientInstance,
                    transport: 'Websocket',
                    transportServiceId: transportServiceId
                } as GeneralEvent<ClientObject>)
                // Update connected clientInstance info to adapter
                connectedClientSocket.push(clientInstance)
                addClientToDB(clientInstance, `clients`)
                startListening(socket, clientInstance, event, transportServiceId)
            }

        })
    })
}


// this is for server usage only
export function startListening(socket: SocketForConnectedClient, client: ConnectedSocketClient, eventListener: Observer<GeneralEvent<any>>, transportServiceid: string, oldClient?: boolean): void {
    // notify it's associated retransmission to start releaseing buffer
    eventListener.next({
        id: uuidv4(),
        type: 'Transport Event',
        event: oldClient ? `Client Re-connected` : `Client Connected`,
        date: new Date(),
        data: client,
        transport: 'Websocket',
        transportServiceid: transportServiceid
    } as unknown as GeneralEvent<ClientObject>)

    /* Generally, we don't need this unless in the case of being the receiver */
    socket.on('message', (message: TransportMessage) => {
        console.log({ message: `Message from client ${client.clientId}`, details: message })
        eventListener.next({
            id: uuidv4(),
            type: 'Transport Event',
            event: 'New Message',
            date: new Date(),
            data: message,
            transport: `Websocket`,
            transportServiceId: transportServiceid
        } as GeneralEvent<TransportMessage>)
    })

    socket.on('disconnect', () => {
        eventListener.next({
            id: uuidv4(),
            type: `Transport Event`,
            event: 'Client Disconnected',
            date: new Date(),
            data: client,
            transport: 'Websocket',
            transportServiceId: transportServiceid
        } as GeneralEvent<ClientObject>)
        eventListener.error(`Client ${client.clientId} disconnected. Terminating this observable event for this client socket...`)
        eventListener.complete()
    })
}
